"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from __future__ import annotations
from enum import Enum

from loom.eka.utilities.enums import enum_missing


class FourBodySchedule(str, Enum):
    """
    Enum for the four-body schedule used in the rotated surface code.
    """

    N = "N"
    Z = "Z"

    def opposite_schedule(self) -> FourBodySchedule:
        """Get the opposite schedule."""
        if self == FourBodySchedule.N:
            return FourBodySchedule.Z
        if self == FourBodySchedule.Z:
            return FourBodySchedule.N
        raise ValueError("Invalid schedule. Cannot determine opposite schedule.")

    @classmethod
    def _missing_(cls, value):
        """
        Allow inputs with upper-case characters. For more details, see the
        documentation of ``enum_missing`` at the beginning of the file.
        """
        return enum_missing(cls, value)
